#include "crefl.1.7.1_viirs.h"
#include <getopt.h>
#include <math.h>
#include <hdf5.h>

static void filter_viirs_float_data (float32 *data, int nr_rows, int nr_cols, float32 *fill_value);
static void filter_viirs_short_data (uint16  *data, int nr_rows, int nr_cols, uint16  *fill_value);

enum { OPT_RANGE = 1, OPT_OUTFILE, OPT_MAXSOLZ };

void viirs_process_options (int main_argc, char *main_argv [], crefl_in_options_st *options_val_st, char **filename, unsigned char *process)
{
    int c, j, ftype;
    int option_index = 0;
    
    struct option long_options [] = {
		{"append", no_argument, &(options_val_st->append), 1},
		{"gzip", no_argument, &(options_val_st->gzip), 1},
		{"maxsolz", required_argument, (int *) NULL, OPT_MAXSOLZ},
		{"nearest", no_argument, &(options_val_st->nearest), 1},
		{"of", required_argument, (int *) NULL, OPT_OUTFILE},
		{"overwrite", no_argument, &(options_val_st->overwrite), 1},
		{"range", required_argument, (int *) NULL, OPT_RANGE},
		{"sealevel", no_argument, &(options_val_st->sealevel), 1},
		{"toa", no_argument, &(options_val_st->TOA), 1},
		{"verbose", no_argument, &(options_val_st->verbose), 1},
		{(char *) NULL, 0, (int *) NULL, 0}
	};
    
	/* Subtract main_argc with one, and start from the second element given that the first element is 
	 * the sensor parameter value */
	while ((c = getopt_long (main_argc, main_argv, "", long_options, &option_index)) >= 0) {

		switch (c) {
		case 0:
			/* do nothing for options which will have a
			 * flag set automatically by getopt_long() */
			break;

		case OPT_RANGE:
			if (sscanf(optarg, "%g,%g", &(options_val_st->reflmin), &(options_val_st->reflmax)) != 2) {
				fputs("Error parsing reflectance range.\n",
				    stderr);
				exit(1);
			}

			if (range_check(options_val_st->reflmin, 0.0F, 1.0F) ||
			    range_check(options_val_st->reflmax, 0.0F, 1.0F)
			    || (options_val_st->reflmin >= options_val_st->reflmax)) {
				fputs("Invalid reflectance range.\n", stderr);
				exit(1);
			}

			printf
			    ("Output reflectance range [%.3f,%.3f] requested.\n",
			    options_val_st->reflmin, options_val_st->reflmax);
			break;

		case OPT_MAXSOLZ:
			options_val_st->maxsolz = (float) atof(optarg);
			if (range_check(options_val_st->maxsolz, 0.0F, 90.0F)) {
				fputs("Invalid max. solar zenith angle.\n",
				    stderr);
				exit(1);
			}
			break;

		case OPT_OUTFILE:
			*filename = optarg;
			break;

		default:
			viirs_usage();
			exit(1);
		}
	}
    
	/* at least one input file must follow */
	if (optind >= main_argc) {
		viirs_usage();
		exit(1);
	}

	/* check for conflicting options */
	if (options_val_st->overwrite && options_val_st->append) {
		fputs
		    ("Options --overwrite and --append are mutually exclusive.\n",
		    stderr);
		exit(1);
	}
	if (options_val_st->sealevel && options_val_st->TOA) {
		fputs("Options --sealevel and --toa are mutually exclusive.\n",
		    stderr);
		exit(1);
	}
#ifdef DEBUG
	printf("append = %d\n", options_val_st->append);
	if (*filename)
		printf("output filename = %s\n", *filename);
	printf("gzip = %d\n", options_val_st->gzip);
	printf("nearest = %d\n", options_val_st->nearest);
	printf("sealevel = %d\n", options_val_st->sealevel);
	printf("TOA = %d\n", options_val_st->TOA);
	printf("Max. solar zenith angle: %g degrees\n", options_val_st->maxsolz);
	if (*filename)
		printf("Output file: %s.\n", *filename);
#endif
	if (options_val_st->verbose)
		printf("Verbose mode requested.\n");
	if (options_val_st->overwrite)
		printf("Overwriting existing output file.\n");
	if (options_val_st->gzip)
		printf("Gzip compression requested.\n");
	if (options_val_st->sealevel)
		printf("Sea-level atmospheric correction requested. Terrain height ignored.\n");
	if (options_val_st->TOA)
		printf("Top-of-the-atmosphere reflectance requested. No atmospheric correction.\n");
	if (options_val_st->nearest)
		printf("Interpolation disabled.\n");
    
    for (j = optind; j < main_argc; j++) {
        ftype = viirs_input_file_type (main_argv [j]);
        
        if (ftype == INPUT_UNKNOWN) {
            fprintf (stderr, "Unrecognized input file \"%s\".\n", main_argv [j]);
            exit (1);
        }
        
        options_val_st->input_files [ftype] = main_argv [j];
        if ((ftype != VIIRS_GMTCO) && (ftype != VIIRS_GMODO) && (ftype != VIIRS_GITCO)) {
            process [ftype] = TRUE;
        }

        #ifdef DEBUG
        switch (ftype) {
        case VIIRS_I1:
            printf ("SVI01file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_I2:
            printf ("SVI02file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M7:
            printf ("SVM07file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M3:
            printf ("SVM03file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M4:
            printf ("SVM04file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M8:
            printf ("SVM08file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_I3:
            printf ("SVI03file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M10:
            printf ("SVM10file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M11:
            printf ("SVM11file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M5:
            printf ("SVM05file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_M2:
            printf ("SVM02file = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_GMTCO:
            printf ("GMTCOfile = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_GMODO:
            printf ("GMODOfile = %s\n", options_val_st->input_files [ftype]);
            break;
        case VIIRS_GITCO:
            printf ("GITCOfile = %s\n", options_val_st->input_files [ftype]);
            break;
        }
        #endif /* DEBUG */
    }
    
    /* Double check that there the GMTCO or GMODO file name has been given */
    if ((options_val_st->input_files [VIIRS_GMTCO] == NULL) && (options_val_st->input_files [VIIRS_GMODO] == NULL) && (options_val_st->input_files [VIIRS_GITCO] == NULL)) {
        fputs("Missing VIIRS GMTCO or GMODO or GITCO file.\n", stderr);
        exit (1);
    }
    
	/* output file name is mandatory */
	if (!(*filename)) {
		fputs("Missing output file name.\n", stderr);
		exit(1);
	}
}

static char *viirs_data_group_path_str_arr [NR_FILES] = {
    "/All_Data/VIIRS-I1-SDR_All",
    "/All_Data/VIIRS-I2-SDR_All",
    "/All_Data/VIIRS-M7-SDR_All",
    "/All_Data/VIIRS-M3-SDR_All",
    "/All_Data/VIIRS-M4-SDR_All",
    "/All_Data/VIIRS-M8-SDR_All",
    "/All_Data/VIIRS-I3-SDR_All",
    "/All_Data/VIIRS-M10-SDR_All",
    "/All_Data/VIIRS-M11-SDR_All",
    "/All_Data/VIIRS-M5-SDR_All",
    "/All_Data/VIIRS-M2-SDR_All",
    "/All_Data/VIIRS-MOD-GEO-TC_All",
    "/All_Data/VIIRS-MOD-GEO_All",
    "/All_Data/VIIRS-IMG-GEO-TC_All"
};

static int viirs_mapping_to_modis_band_nr_arr [NR_FILES] = {
    0, /* I1 */  /* MODIS BAND 1 */ /* 640 nm  */ /* 645.5  nm */
    1, /* I2 */  /* MODIS BAND 2 */ /* 865 nm  */ /* 856.5  nm */
    1, /* M7 */  /* MODIS BAND 2 */ /* 865 nm  */ /* 856.5  nm */
    2, /* M3 */  /* MODIS BAND 3 */ /* 488 nm  */ /* 465.6  nm */
    3, /* M4 */  /* MODIS BAND 4 */ /* 555 nm  */ /* 553.6  nm */
    4, /* M8 */  /* MODIS BAND 5 */ /* 1024 nm */ /* 1241.6 nm */
    5, /* I3 */  /* MODIS BAND 6 */ /* 1610 nm */ /* 1629.1 nm */
    5, /* M10 */ /* MODIS BAND 6 */ /* 1610 nm */ /* 1629.1 nm */
    6, /* M11 */ /* MODIS BAND 7 */ /* 2250 nm */ /* 2114.1 nm */
    0, /* M5  */ /* MODIS BAND 1 */ /* 672  nm */ /* 645.5  nm */
    2, /* M2 */  /* MODIS BAND 3 */ /* 445  nm */ /* 465.6  nm */
    0, /* fill value - will not be used */
    0, /* fill value - will not be used */
    0  /* fill value - will not be used */
};

static int viirs_sensor_band_nr_arr [NR_FILES] = {
    11, /* I1 */
    6,  /* I2 */
    6,  /* M7 */
    2,  /* M3 */
    3,  /* M4 */
    7,  /* M8 */
    9,  /* I3 */
    9,  /* M10 */
    10, /* M11 */
    4,  /* M5  */
    1,  /* M2 */
    0,  /* fill value - will not be used */
    0,  /* fill value - will not be used */
    0   /* fill value - will not be used */
};

void viirs_init_data (crefl_in_options_st *options_val_st, SDS *sds_st_arr, unsigned char *process)
{
    int idx, file_idx;
    hid_t dataset_data_type;
    hsize_t dim_arr [2], dim_sub_arr [2];
    hid_t refl_fac_did;
    size_t data_size;
    char data_set_path_str [255];
    float32 scale_offset_arr [2];
    char bool_geo_refl_data; /* 1 -> geo data, 0 -> refl data */
    
    for (idx = 0; idx < Nitems; ++idx) {
        if (idx < SOLZ) {
            if (process [idx] == FALSE) {
                sds_st_arr [idx].id = -1;
                continue;
            }

            file_idx = idx;
            
            bool_geo_refl_data = 0;
        } else {
            if (options_val_st->input_files [VIIRS_GMTCO] != NULL) {
                file_idx = VIIRS_GMTCO;
            }
            
            if (options_val_st->input_files [VIIRS_GMODO] != NULL) {
                file_idx = VIIRS_GMODO;
            }
            
            if (options_val_st->input_files [VIIRS_GITCO] != NULL) {
                file_idx = VIIRS_GITCO;
            }
            
            bool_geo_refl_data = 1;
        }
        
        sds_st_arr [idx].id = 0;
        
        /* TODO: For the geolocation data, there will be multiple opens for one file. Find a way arround it. */
        sds_st_arr [idx].h5_file_id = H5Fopen (options_val_st->input_files [file_idx], H5F_ACC_RDONLY, H5P_DEFAULT);
        if (sds_st_arr [idx].h5_file_id < 0) {
            fprintf (stderr, "Cannot open input file %s.\n", options_val_st->input_files [file_idx]);
            exit (1);
        }
        
        /* Set the data set path for this specfic band, so that it can be opened */
        
        if (idx < SOLZ) {
            sds_st_arr [idx].name = viirs_data_group_path_str_arr [idx];
            
            strcpy (data_set_path_str, viirs_data_group_path_str_arr [idx]);
            strcat (data_set_path_str, "/Reflectance");
        } else {
            sds_st_arr [idx].name = viirs_data_group_path_str_arr [file_idx];
            
            strcpy (data_set_path_str, viirs_data_group_path_str_arr [file_idx]);
            switch (idx) {
                case SOLZ:
                    strcat (data_set_path_str, "/SolarZenithAngle");
                break;
                
                case SENZ:
                    strcat (data_set_path_str, "/SatelliteZenithAngle");
                break;
                
                case SOLA:
                    strcat (data_set_path_str, "/SolarAzimuthAngle");
                break;
                
                case SENA:
                    strcat (data_set_path_str, "/SatelliteAzimuthAngle");
                break;
                
                case LON:
                    strcat (data_set_path_str, "/Longitude");
                break;
                
                case LAT:
                    strcat (data_set_path_str, "/Latitude");
                break;
            }
        }
        
        #if (H5_VERS_MINOR <= 6) && (H5_VERS_MAJOR == 1)
        /* HDF5 version <= 1.6 */
        sds_st_arr [idx].h5_dataset_id = H5Dopen (sds_st_arr [idx].h5_file_id, data_set_path_str);
        #else
        /* HDF5 version >= 1.7 */
        sds_st_arr [idx].h5_dataset_id = H5Dopen (sds_st_arr [idx].h5_file_id, data_set_path_str, H5P_DEFAULT);
        #endif
        
        if (sds_st_arr [idx].h5_dataset_id < 0) {
            fprintf (stderr, "Cannot open dataset %s.\n", data_set_path_str);
            exit (1);
        }
        
        /* Get the dataset type and the dataset type size */
        dataset_data_type = H5Dget_type (sds_st_arr [idx].h5_dataset_id);
        if (dataset_data_type < 0) {
            fprintf (stderr, "Cannot get data type for dataset %s.\n", data_set_path_str);
            exit (1);
        }
        data_size = H5Tget_size (dataset_data_type);
        H5Tclose (dataset_data_type);

        /* Get the dimensions of the dataset */
        sds_st_arr [idx].h5_dataspace_id = H5Dget_space (sds_st_arr [idx].h5_dataset_id);
        sds_st_arr [idx].rank = H5Sget_simple_extent_ndims (sds_st_arr [idx].h5_dataspace_id);
        if (H5Sget_simple_extent_dims (sds_st_arr [idx].h5_dataspace_id, dim_arr, NULL) < 0) {
            fprintf (stderr, "Could not determine the dimensions for the dataset %s.\n", data_set_path_str);
            exit (1);
        }
        sds_st_arr [idx].Nl = dim_arr [0];
        sds_st_arr [idx].Np = dim_arr [1];
        
        /* Calculate the number rows per scan */
        sds_st_arr [idx].rowsperscan =
			    (int) (VIIRS_NUM750MROWPERSCAN * sds_st_arr [idx].Np / (float) VIIRS_NUM750MCOLPERSCAN + 0.5);
        
        /* Set the variables for the hyperslab function */
        dim_sub_arr [0] = sds_st_arr [idx].rowsperscan;
        dim_sub_arr [1] = sds_st_arr [idx].Np;
        sds_st_arr [idx].h5_memspace_id = H5Screate_simple (sds_st_arr [idx].rank, dim_sub_arr, NULL);
        
        /* Allocate memory for the fillvalue variable and set it */
        sds_st_arr [idx].fillvalue = (void *) malloc (data_size);
        if (bool_geo_refl_data == 0) {
            *((uint16 *)sds_st_arr [idx].fillvalue) = 65535;
        } else {
            *((float32 *)sds_st_arr [idx].fillvalue) = -999.0;
        }
        
        /* Allocate memory for the data variable */
        sds_st_arr [idx].data = (void *) malloc (sds_st_arr [idx].Np * sds_st_arr [idx].rowsperscan * data_size);
        
        if (bool_geo_refl_data == 0) {
            sds_st_arr [idx].h5_datatype = H5T_NATIVE_UINT16;
        } else {
            sds_st_arr [idx].h5_datatype = H5T_NATIVE_FLOAT;
        }
        
        sds_st_arr [idx].h5_start [1] = 0;
        sds_st_arr [idx].h5_stride [0] = 1;
        sds_st_arr [idx].h5_stride [1] = 1;
        sds_st_arr [idx].h5_count [0] = sds_st_arr [idx].rowsperscan;
        sds_st_arr [idx].h5_count [1] = sds_st_arr [idx].Np;
        
        if (bool_geo_refl_data == 0) {
            sds_st_arr [idx].crefl_band_number = viirs_mapping_to_modis_band_nr_arr [idx];
            sds_st_arr [idx].sensor_band_number = viirs_sensor_band_nr_arr [idx];
            
            /* Get the factor and offset values */
            strcpy (data_set_path_str, viirs_data_group_path_str_arr [idx]);
            strcat (data_set_path_str, "/ReflectanceFactors");
            
            #if (H5_VERS_MINOR <= 6) && (H5_VERS_MAJOR == 1)
            /* HDF5 version <= 1.6 */
            refl_fac_did = H5Dopen (sds_st_arr [idx].h5_file_id, data_set_path_str);
            #else
            /* HDF5 version >= 1.7 */;
            refl_fac_did = H5Dopen (sds_st_arr [idx].h5_file_id, data_set_path_str, H5P_DEFAULT);
            #endif
            
            if (H5Dread (refl_fac_did, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, scale_offset_arr) < 0) {
                fprintf (stderr, "Could not read the data of dataset %s.\n", data_set_path_str);
            }
            sds_st_arr [idx].factor = scale_offset_arr [0];
            sds_st_arr [idx].offset = scale_offset_arr [1];
            H5Dclose (refl_fac_did);
        } else {
            sds_st_arr [idx].factor = 1.0;
            sds_st_arr [idx].offset = 0.0;
        }
    }
}

void viirs_finalize_data (crefl_in_options_st *options_val_st, SDS *sds_st_arr)
{
    int idx;
    
    for (idx = 0; idx < Nitems; idx++) {
        if (sds_st_arr [idx].id == -1) {
            continue;
        }
        
        free (sds_st_arr [idx].data);
        free (sds_st_arr [idx].fillvalue);
        
        H5Sclose (sds_st_arr [idx].h5_dataspace_id);
        H5Sclose (sds_st_arr [idx].h5_memspace_id);
        H5Dclose (sds_st_arr [idx].h5_dataset_id);
        H5Fclose (sds_st_arr [idx].h5_file_id);
    }
}

void viirs_usage (void)
{
	fputs("Usage for VIIRS data:\n", stderr);
	fputs
	    ("crefl viirs [--verbose] [--nearest] [--toa|--sealevel]\n"
	    "      [--gzip] [--maxsolz=angle] [--range=min,max] [--overwrite|--append]\n"
	    "       --of=<output file>\n"
        "      <GMTCO | GMODO | GITCO file> [<SVI01 file>] [<SVI02 file>] [<SVM07 file>] [<SVM02 file>] [<SVM04 file>]\n"
        "      [<SVM08 file>] [<SVI03 file>] [<SVM10 file>] [<SVM11 file>] [<SVM05 file>] [<SVM02 file>]\n", 
	    stderr);

	fprintf(stderr, "Version %s, compiled %s %s.\n", PROCESS_VERSION_NUMBER, __DATE__, __TIME__);
}

int viirs_input_file_type (char *filename)
{
    char *ptr;
	int maxlen;

	if ((ptr = strrchr(filename, '/')))
		/* move to first character after last forward slash */
		ptr++;
	else
		/* no slash in file name */
		ptr = filename;

	/* all prefixes we'll recognize have this length */
	maxlen = strlen("GMTCO_");

	if (strncmp(ptr, "GMTCO_", maxlen) == 0)
		return VIIRS_GMTCO;
    if (strncmp(ptr, "GMODO_", maxlen) == 0)
        return VIIRS_GMODO;
    if (strncmp(ptr, "GITCO_", maxlen) == 0)
        return VIIRS_GITCO;
	if (strncmp(ptr, "SVI01_", maxlen) == 0)
		return VIIRS_I1;
	if (strncmp(ptr, "SVI02_", maxlen) == 0)
		return VIIRS_I2;
	if (strncmp(ptr, "SVI03_", maxlen) == 0)
		return VIIRS_I3;
	if (strncmp(ptr, "SVM03_", maxlen) == 0)
		return VIIRS_M3;
	if (strncmp(ptr, "SVM04_", maxlen) == 0)
		return VIIRS_M4;
	if (strncmp(ptr, "SVM07_", maxlen) == 0)
		return VIIRS_M7;
	if (strncmp(ptr, "SVM08_", maxlen) == 0)
		return VIIRS_M8;
	if (strncmp(ptr, "SVM10_", maxlen) == 0)
		return VIIRS_M10;
	if (strncmp(ptr, "SVM11_", maxlen) == 0)
		return VIIRS_M11;
    if (strncmp(ptr, "SVM05_", maxlen) == 0)
		return VIIRS_M5;
    if (strncmp(ptr, "SVM02_", maxlen) == 0)
		return VIIRS_M2;

	return INPUT_UNKNOWN;
}

int viirs_clause_solz_neg_l1bdata (SDS *solz_st, SDS *l1bdata_st, SDS *outsds_st, int crsidx, int idx)
{
    float32 *solz = (float32 *)solz_st->data;
    float32 *solzfill = (float32 *)solz_st->fillvalue;
    uint16 *l1bdata = (uint16 *)l1bdata_st->data;
    uint16 *l1b_fillvalue = (uint16 *)l1bdata_st->fillvalue;
    int16 *outdata = (int16 *)outsds_st->data;
    int16 *outdata_fill = (int16 *)outsds_st->fillvalue;
    
    if ((solz [crsidx] == *solzfill) || (l1bdata [idx] == *l1b_fillvalue)) {
        outdata [idx] = *outdata_fill;
        
        return 1;
    }

    return 0;
}

int viirs_test_solz_gt_maxsolz (SDS *solz_st, int idx, float maxsolz)
{
    float32 *solz = solz_st->data;
    
    if ((solz [idx] * solz_st->factor) >= maxsolz) {
        return 1;
    }

    return 0;
}

int viirs_test_solz_ne_fillvalue (SDS *solz_st, int idx)
{
    float32 *solz = solz_st->data;
    float32 *solzfill = solz_st->fillvalue;

    if (solz [idx] != *solzfill) {
        return 1;
    }

    return 0;
}

int viirs_test_bad_val (SDS *solz_st, int crsidx11, int crsidx12, int crsidx21, int crsidx22)
{
    int bad;
    float32 *solz = solz_st->data;
    float32 *solzfill = solz_st->fillvalue;
    
    bad = (solz [crsidx11] == *solzfill) || (solz [crsidx12] == *solzfill) || (solz [crsidx21] == *solzfill) || (solz [crsidx22] == *solzfill);
    
    return bad;
}

void viirs_set_solz_to_fillvalue (SDS *solz_st, int idx)
{
    float32 *solz = solz_st->data;
    float32 *solzfill = solz_st->fillvalue;
    
    solz [idx] = *solzfill;
}

float viirs_calculate_refl (SDS *l1bdata_st, int idx)
{
    float refl;
    uint16 *l1bdata = l1bdata_st->data;
    
    refl = (l1bdata [idx] - l1bdata_st->offset) * l1bdata_st->factor;

    return refl;
}

float viirs_calculate_mus (SDS *solz_st, int idx)
{
    float mus;
    float32 *solz = solz_st->data;
    
    mus = cos (solz [idx] * solz_st->factor * DEG2RAD);
    
    return mus;
}

float viirs_calculate_phi (SDS *sola_st, SDS *sena_st, int idx)
{
    float phi;
    float32 *sola = sola_st->data;
    float32 *sena = sena_st->data;
    
    phi = sola [idx] * sola_st->factor - sena [idx] * sena_st->factor;
    
    return phi;
}

float viirs_calculate_muv (SDS *senz_st, int idx)
{
    float muv;
    float32 *senz = senz_st->data;
    
    muv = cos (senz [idx] * senz_st->factor * DEG2RAD);
    
    return muv;
}

int viirs_read_scan (int iscan, SDS sds_st_arr [Nitems])
{
    int ib;

	for (ib = 0; ib < Nitems; ib++) {
		if (sds_st_arr[ib].id == -1)
			continue;
        
	    sds_st_arr[ib].h5_start [0] = iscan * sds_st_arr [ib].rowsperscan;
        
        H5Sselect_hyperslab (sds_st_arr [ib].h5_dataspace_id, H5S_SELECT_SET, sds_st_arr [ib].h5_start, 
            sds_st_arr [ib].h5_stride, sds_st_arr [ib].h5_count, NULL);
        
		if (H5Dread (sds_st_arr[ib].h5_dataset_id, sds_st_arr [ib].h5_datatype, sds_st_arr [ib].h5_memspace_id, 
                sds_st_arr[ib].h5_dataspace_id, H5P_DEFAULT, sds_st_arr [ib].data) < 0) {
            
			fprintf(stderr, "  Can't read scan %d of dataset \"%s\"\n", iscan, sds_st_arr[ib].name);
			return 1;
		}
        
        if (sds_st_arr [ib].h5_datatype == H5T_NATIVE_FLOAT) {
            filter_viirs_float_data (sds_st_arr [ib].data, sds_st_arr [ib].rowsperscan, sds_st_arr [ib].Np, (float32 *)sds_st_arr [ib].fillvalue);
        } else {
            filter_viirs_short_data (sds_st_arr [ib].data, sds_st_arr [ib].rowsperscan, sds_st_arr [ib].Np, (uint16 *)sds_st_arr [ib].fillvalue);
        }
	}

	return 0;
}

void filter_viirs_float_data (float32 *data, int nr_rows, int nr_cols, float32 *fill_value)
{
    int idx;
    
    for (idx = 0; idx < nr_rows * nr_cols; ++idx) {
        if (data [idx] < -900.0)
            data [idx] = *fill_value;
    }
}

static void filter_viirs_short_data (uint16 *data, int nr_rows, int nr_cols, uint16 *fill_value)
{
    int idx;
    
    for (idx = 0; idx < nr_rows * nr_cols; ++idx) {
        if (data [idx] > 65527)
            data [idx] = *fill_value;
    }
}

