/*
Linux
  cc -O crefl.c -o crefl -I$HDFINC -L$HDFLIB -lmfhdf -ldf -lz -lm -ljpeg
*/

/*************************************************************************
Description:

  Simplified atmospheric correction algorithm that transforms MODIS
  top-of-the-atmosphere level-1B radiance data into corrected reflectance
  for Rapid Response applications.
  Required ancillary data: coarse resolution DEM tbase.hdf

References and Credits:

  Jacques Descloitres, MODIS Rapid Response Project, NASA/GSFC/SSAI
  http://rapidfire.sci.gsfc.nasa.gov

Revision history:

  Version 1.0   08/24/01
  Version 1.1   01/25/02
  Version 1.2   05/30/02
  Version 1.3   09/06/02
  Version 1.4   09/02/03
  Version 1.4.1 01/22/04
  Version 1.5   02/17/04
  Version 1.5.1 03/08/07 (not run in RR production)
  Version 1.5.2 06/07/07
  Version 1.6   08/18/09 (Be sure to update PROCESS_VERSION_NUMBER also)
                         Started with Version 1.5.2, version run in Rapid Response for several years,
                         this had several differences from 1.4.2, the DRL version, including the addition
                         of band 8 and a number of small, but perhaps important, computational changes
                         1) removed most code within #ifdef DEBUG clauses
                         2) left command line options for nearest, TOA, and sealevel; note that these options
                            were in Version 1.4.2 but were not available from the command line
                         3) changes by DRL to 1.4.2 to write scale factor and offset has already been 
                            incorporated into 1.5.2
                         4) Added in the modifications from Chuanmin Hu & Brock Murch of Univ South Florida
                            IMaRS to add bands 9-16.  The aO3 and taur0 parameters came "from SeaDAS codes"
                            and "The H2O parameters can be assumed 0 for bands 8-16 because those bands were
                            designed to avoid water vapor absorption."
                         Disclaimer: the nearest, TOA, and sealevel options and bands 9-16 are not used
                                     by Rapid Response so I cannot consider them tested/validated - JES

   SSEC, Willem.Marais@ssec.wisc.edu - Modifications:
   --------------------------------------------------
   
   05/20/2013 - TODO
   - Refactor the code. In most places it is messy.
   
   05/20/2013
   - Modified the code so that new atmospheric corrected coefficients are used on VIIRS data. The code was taken 
   from Eric Vermote's VIIRS CREFL code, released by DRL. Refer to the forum post 
   https://forums.ssec.wisc.edu/viewtopic.php?f=42&t=304.
   - The function getatmvariables was removed. The functions modis_getatmvariables and viirs_getatmvariables 
   were added. 
   
   01/16/2013 - TODO
   - Change the crefl output SDS array names to something more meaningful, for VIIRS at least. For example 
   rather than having CorrRefl_7, let the output SDS array name be CorrRefl_I2 or CorrRefl_M7.
   
   10/09/2012 - TODO
   - Cleanup memory allocations after any failure.
   
   07/12/2012
   - The code can now process both VIIRS and MODIS data
   - VIIRS input data are in HDF5 format. MODIS data are in HDF4 format
   - HDF5 library version 1.6.10 is used
   - For VIIRS: Either GMTCO, GMODO or GITCO geolocation files can be used
   
   10/27/2011
   - Adapted code for HDF4 version 4.2.6
   - The code compiles on Red Hat Enterprise 6.1. The jpeg library had to be 
     compiled. jpeg-8c of http://www.ijg.org was used. 
   - indent -bap -br -ce -ci4 -cli0 -d0 -di0 -i8 -ip4 -l79 -nbc -ncdb -ndj -nfc1 -nlp -npcs -psl -sc -sob
   - Remove unnecessary variables 
   - Added notes

   10/31/2011 - Notes
   - The options --1km and --500m are output parameter settings
   - The following variables are shared between initializing code, and the processing code:
       * append: This is an option parameter.
       * dem: SDS structure.
       * filename: Output file name.
       * gzip: This is an option parameter.
       * ib: IndexBand number variable
       * maxsolz: This is an option parameter.
       * MOD021KMfile: The MOD02 1KM filename.
       * MOD02HKMfile: The MOD02 HKM filename.
       * MOD02QKMfile: The MOD02 QKM filename.
       * MOD021KMfile_id: The MOD02 1KM file id returned by SDstart.
       * MOD02HKMfile_id: The MOD02 HKM file id returend by SDstart.
       * MOD02QKMfile_id: The MOD02 QKM file id returned by SDstart.
       * nearest: This is an option parameter.
       * Nscans: The number of scans in the L1B file.
       * process: An array with Nbands number of elements. Each elements 
                  is either equal to TRUE or FALSE.
       * reflmax, reflmin: Reflection maximum and minimum, set by the range 
                           input option parameter.
       * sds: An SDS structure array, with Nitems array elements.
       * sealevel: This is an option parameter.
       * TOA: This is an option parameter.
       * verbose: This is an option parameter.
       * write_mode: Affected by the variable append.
   - Added option: viirs, modis. This is to indicate what sensor's data 
                   will be processed.
*************************************************************************/

#include "crefl.1.7.1_modis.h"
#include "crefl.1.7.1_viirs.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <getopt.h>

#define UO3 0.319
#define UH2O 2.93

#define REFSDS SOLZ

#define MAXSOLZ 86.5
#define MAXAIRMASS 18
#define SCALEHEIGHT 8000
#define FILL_INT16  32767
/* Willem.Marais@ssec.wisc.edu - 10/09/2012 - BEGIN */
#define FILL_FLOAT32 -999
/* Willem.Marais@ssec.wisc.edu - 10/09/2012 - END */

#define TAUSTEP4SPHALB      0.0001
#define MAXNUMSPHALBVALUES  4000    /* with no aerosol taur <= 0.4 in all bands everywhere */

/* Willem.Marais@ssec.wisc.edu - New variables - BEGIN */
enum {SENSOR_MODIS, SENSOR_VIIRS, SENSOR_UNKOWN};
char *sensor_names [2] = {"modis", "viirs"};
/* Willem.Marais@ssec.wisc.edu - New variables - END */
/* Willem.Marais@ssec.wisc.edu - New functions - BEGIN */
char sensor_name_parm (char *option_str);
/* Willem.Marais@ssec.wisc.edu - New functions - END */

void set_dimnames(int samples, char **dimname1, char **dimname2);

int init_output_sds(int32 sd_id, unsigned char *process,
    SDS outsds[Nbands], SDS sds_st_arr[Nitems], int gzip, int verbose);

/* Willem.Marais@ssec.wisc.edu - 10/09/2012 - BEGIN */
int init_geo_output_sds(int32 sd_id, SDS geo_outsds[2], SDS sds_st_arr[Nitems], int gzip, int verbose);
/* Willem.Marais@ssec.wisc.edu - 10/09/2012 - END */

int write_scan_outsds(int iscan, unsigned char *process, SDS outsds[Nbands]);
int write_scan_geo_outsds(int iscan, SDS geo_outsds[2]);
int write_global_attributes(int32 sd_id, char *MOD021KMfile,
    char *MOD02HKMfile, char *MOD02QKMfile,
    float maxsolz, int sealevel, int TOA, int nearest);

/* Willem.Marais@ssec.wisc.edu - 05/20/2013 - BEGIN */
int modis_getatmvariables(float mus, float muv, float phi, int16 height,
    unsigned char *process, float *sphalb, float *rhoray,
    float *TtotraytH2O, float *tOG);

int viirs_getatmvariables(float mus, float muv, float phi, int16 height,
    unsigned char *process, float *sphalb, float *rhoray,
    float *TtotraytH2O, float *tOG);

/*
int getatmvariables(float mus, float muv, float phi, int16 height,
    unsigned char *process, float *sphalb, float *rhoray,
    float *TtotraytH2O, float *tOG);
*/
/* Willem.Marais@ssec.wisc.edu - 05/20/2013 - END */

void chand(float phi, float muv, float mus, float *tau, float *rhoray, float *trup, float *trdown, unsigned char *process);
float csalbr(float tau);
double fintexp1(float tau);
double fintexp3(float tau);
float correctedrefl(float refl, float TtotraytH2O, float tOG, float rhoray, float sphalb);
int interp_dem(float lat, float lon, SDS *dem_sds_st);

int main (int argc, char *argv[])
{
    /* Willem.Marais@ssec.wisc.edu - New variables - BEGIN */
    char process_sensor;
    unsigned char band_process[Nbands];
    /* Willem.Marais@ssec.wisc.edu - New variables - END */
    
    char *filename; /* output file */

    FILE *fp;
    int outfile_exists;
    
    SDS sds_st_arr[Nitems], outsds[Nbands], dem_sds_st, height;
    /* Willem.Marais@ssec.wisc.edu - 10/09/2012 - BEGIN */
    SDS geo_outsds[2];
    /* Willem.Marais@ssec.wisc.edu - 10/09/2012 - END */
    int32 sd_id;

    int ib, band_idx, iscan, Nscans, irow, jcol, idx, crsidx, geo_idx;
    
    float32 *lon, *lat, *lonfill, *latfill;
    
    float refl, *mus, muv, phi;
    float *rhoray, *sphalb, *TtotraytH2O, *tOG;
    /* Willem.Marais@ssec.wisc.edu - Modification - BEGIN */
    /* aggfactor was an int, now is a float */
    float aggfactor;
    int crsrow1, crsrow2, crscol1, crscol2;
    /* Willem.Marais@ssec.wisc.edu - Modification - END */
    int crsidx11, crsidx12, crsidx21, crsidx22;
    float mus0;
    float fractrow, fractcol, t, u;
    float rhoray0, rhoray11, rhoray12, rhoray21, rhoray22;
    float sphalb0, sphalb11, sphalb12, sphalb21, sphalb22;
    int bad;

    int write_mode = DFACC_CREATE;
    
    int pixels_per_rowscan;

    size_t nbytes;
    
    enum { OPT_BANDS = 1, OPT_RANGE, OPT_OUTFILE, OPT_MAXSOLZ };

    /* Willem.Marais@ssec.wisc.edu - New code - BEGIN */
    static crefl_in_options_st crefl_in_options_st;
    /* Willem.Marais@ssec.wisc.edu - New code - END */

    filename = (char *) NULL;

    /* Willem.Marais@ssec.wisc.edu - New code - BEGIN */
    /* Check what sensor's data will be processed */
    if (argc < 2) {
        fputs ("Could not determine what sensor's data will be processed.\n", stderr);
        fputs ("\n", stderr);
        modis_usage ();
        fputs ("\n", stderr);
        viirs_usage ();
        
        exit(1);
    }
    process_sensor = sensor_name_parm (argv [1]);
    
    init_crefl_in_options_st (&crefl_in_options_st);
    
    switch (process_sensor) {
        case SENSOR_MODIS:
            modis_process_options (argc - 1, &argv[1], &crefl_in_options_st, &filename, crefl_in_options_st.process);
            pixels_per_rowscan = MODIS_NUM1KMROWPERSCAN;
        break;
        
        case SENSOR_VIIRS:
            viirs_process_options (argc - 1, &argv[1], &crefl_in_options_st, &filename, crefl_in_options_st.process);
            pixels_per_rowscan = VIIRS_NUM750MROWPERSCAN;
        break;
        
        default:
            fputs ("Could not determine what sensor's data will be processed.\n", stderr);
            usage ();
            exit(1);
        break;
    }
    /* Willem.Marais@ssec.wisc.edu - New code - END */
    
    if (crefl_in_options_st.append) {
        write_mode = DFACC_RDWR;
    }
    
    /* ======================================================================================================= */
    /* Open DEM HDF input file */
    /* ======================================================================================================= */
    dem_init_data (&crefl_in_options_st, &dem_sds_st);
    
    /* ======================================================================================================= */
    /* Get input data */
    /* ======================================================================================================= */
    if (process_sensor == SENSOR_MODIS) {
        modis_init_data (&crefl_in_options_st, sds_st_arr, crefl_in_options_st.process);
    } else {
        viirs_init_data (&crefl_in_options_st, sds_st_arr, crefl_in_options_st.process);
    }

    for (idx = 0; idx < Nbands; ++idx) {
        band_process [idx] = 0;
    }
    for (idx = 0; idx < Nbands; ++idx) {
        if (sds_st_arr [idx].id != -1) {
            band_process [(int)sds_st_arr [idx].crefl_band_number] = 1;
        }
    }

    /* ======================================================================================================= */
    /* Test-open output file */
    /* ======================================================================================================= */
    if ((fp = fopen(filename, "r"))) {
        (void) fclose(fp);
        outfile_exists = 1;
    } else
        outfile_exists = 0;

    if ((write_mode == DFACC_CREATE) && !crefl_in_options_st.overwrite && outfile_exists) {
        fprintf(stderr, "File \"%s\" already exits.\n", filename);
        exit(1);
    }
    
    if (sds_st_arr[SOLZ].id == -1 ||
        sds_st_arr[SOLA].id == -1 ||
        sds_st_arr[SENZ].id == -1 ||
        sds_st_arr[SENA].id == -1 ||
        sds_st_arr[LON].id == -1  ||
        sds_st_arr[LAT].id == -1  || ((dem_sds_st.id == -1) && !crefl_in_options_st.sealevel && !crefl_in_options_st.TOA)) {
        fprintf(stderr,
            "Solar and Sensor angles and DEM are necessary to process granule.\n");
        exit(1);
    }

    if (sds_st_arr[REFSDS].Np != sds_st_arr[SOLZ].Np ||
        sds_st_arr[REFSDS].Np != sds_st_arr[SOLA].Np ||
        sds_st_arr[REFSDS].Np != sds_st_arr[SENZ].Np ||
        sds_st_arr[REFSDS].Np != sds_st_arr[SENA].Np ||
        sds_st_arr[REFSDS].Np != sds_st_arr[LON].Np  || sds_st_arr[REFSDS].Np != sds_st_arr[LAT].Np) {
        fprintf(stderr,
            "Solar and Sensor angles must have identical dimensions.\n");
        exit(1);
    }

    ib = 0;
    while (sds_st_arr[ib].id == -1)
        ib++;
    if (ib >= Nbands) {
        fprintf(stderr, "No L1B SDS can be read successfully.\n");
        exit(1);
    }

    Nscans = sds_st_arr[ib].Nl / sds_st_arr[ib].rowsperscan;
    
    /* ------------------------------------------------------------------------------------------------- */
    /* ---------------------------------------------- CUT ---------------------------------------------- */
    /* ------------------------------------------------------------------------------------------------- */
    
    /* ======================================================================================================= */
    /* Output file section */
    /* ======================================================================================================= */
    /* finally, open output file */
    if ((sd_id = SDstart (filename, write_mode)) == -1) {
        fprintf(stderr, "Cannot open output file %s.\n", filename);
        exit(1);
    }

    if (!crefl_in_options_st.append) {
        if (write_global_attributes (sd_id, crefl_in_options_st.input_files [MODIS_1KM], crefl_in_options_st.input_files [MODIS_HKM],
                                     crefl_in_options_st.input_files [MODIS_QKM], crefl_in_options_st.maxsolz, 
                                     crefl_in_options_st.sealevel, crefl_in_options_st.TOA, crefl_in_options_st.nearest)) {
            fputs("Error writing global attributes.\n", stderr);
            exit(1);
        }
    }
    
    /* create output SDSs and set SDS-specific attributes and dimension names */
    if (init_output_sds (sd_id, crefl_in_options_st.process, outsds, sds_st_arr, crefl_in_options_st.gzip, crefl_in_options_st.verbose)) {
        exit(1);
    }
    
    if (init_geo_output_sds (sd_id, geo_outsds, sds_st_arr, crefl_in_options_st.gzip, crefl_in_options_st.verbose)) {
        exit(1);
    }

    mus =
        (float *) malloc(sds_st_arr[REFSDS].rowsperscan * sds_st_arr[REFSDS].Np *
        sizeof(float));
    height.data =
        (int16 *) malloc(sds_st_arr[REFSDS].rowsperscan * sds_st_arr[REFSDS].Np *
        sizeof(int16));
    if (!mus || !height.data) {
        (void) fputs("Error allocating memory.\n", stderr);
        exit(1);
    }

    if (crefl_in_options_st.sealevel || crefl_in_options_st.TOA)
        dem_sds_st.data = (void *) NULL;
    else {
        dem_sds_st.data = (int16 *) malloc (dem_sds_st.Nl * dem_sds_st.Np * sizeof(int16));
        if (!dem_sds_st.data) {
            (void) fputs("Error allocating memory.\n", stderr);
            exit(1);
        }
    }

    if (!crefl_in_options_st.TOA) {
        nbytes =
            Nbands * sds_st_arr[REFSDS].rowsperscan * sds_st_arr[REFSDS].Np *
            sizeof(float);

        rhoray = (float *) malloc(nbytes);
        sphalb = (float *) malloc(nbytes);
        TtotraytH2O = (float *) malloc(nbytes);
        tOG = (float *) malloc(nbytes);

        if (!rhoray || !sphalb || !TtotraytH2O || !tOG) {
            (void) fputs("Error allocating memory.\n", stderr);
            exit(1);
        }
    }

    lon = sds_st_arr[LON].data;
    lat = sds_st_arr[LAT].data;
    lonfill = sds_st_arr[LON].fillvalue;
    latfill = sds_st_arr[LAT].fillvalue;

    /* don't need DEM if --sealevel or --toa specified */
    if (!crefl_in_options_st.sealevel && !crefl_in_options_st.TOA) {
        dem_sds_st.start[0] = 0;
        dem_sds_st.start[1] = 0;
        dem_sds_st.edges[0] = dem_sds_st.Nl;
        dem_sds_st.edges[1] = dem_sds_st.Np;
        
        if (SDreaddata(dem_sds_st.id, dem_sds_st.start, NULL, dem_sds_st.edges,
            dem_sds_st.data) == -1) {
            fprintf(stderr, "  Can't read DEM SDS \"%s\"\n",
                dem_sds_st.name);
            exit(-1);
        }
        (void) SDendaccess(dem_sds_st.id);
        (void) SDend(dem_sds_st.file_id);
    }
    
    
    /* loop over each MODIS scan */
    for (iscan = 0; iscan < Nscans; iscan++) {
        if ((iscan % pixels_per_rowscan == 0) && crefl_in_options_st.verbose)
            printf("Processing scan %d...\n", iscan);

        /* Fill scan buffer for each band to be processed.
         * Exit scan loop if error occurred while reading. */
        if (process_sensor == SENSOR_MODIS) {
            if (modis_read_scan (iscan, sds_st_arr))
                break;
        } else {
            if (viirs_read_scan (iscan, sds_st_arr))
                break;
        }

        for (idx = 0; idx < sds_st_arr[REFSDS].rowsperscan * sds_st_arr[REFSDS].Np; idx++) {
            if (process_sensor == SENSOR_MODIS) {
                if (modis_test_solz_gt_maxsolz (&sds_st_arr [SOLZ], idx, crefl_in_options_st.maxsolz)) {
                    modis_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                }
            } else {
                if (viirs_test_solz_gt_maxsolz (&sds_st_arr [SOLZ], idx, crefl_in_options_st.maxsolz)) {
                    viirs_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                }
            }
          
            if (!crefl_in_options_st.sealevel && (lon[idx] == *lonfill || lat[idx] == *latfill)) {
                if (process_sensor == SENSOR_MODIS) {
                    modis_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                } else {
                    viirs_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                }
            }
            
            if (process_sensor == SENSOR_MODIS) {
                if (modis_test_solz_ne_fillvalue (&sds_st_arr [SOLZ], idx)) {
                    mus [idx] = modis_calculate_mus (&sds_st_arr [SOLZ], idx);
                    
                    if (crefl_in_options_st.sealevel || crefl_in_options_st.TOA)
                        ((int16 *) height.data)[idx] = 0;
                    else
                        ((int16 *) height.data)[idx] =
                            (int16) interp_dem(lat[idx],
                            lon[idx], &dem_sds_st);
                }
            } else {
                if (viirs_test_solz_ne_fillvalue (&sds_st_arr [SOLZ], idx)) {
                    mus [idx] = viirs_calculate_mus (&sds_st_arr [SOLZ], idx);
                    
                    if (crefl_in_options_st.sealevel || crefl_in_options_st.TOA)
                        ((int16 *) height.data)[idx] = 0;
                    else
                        ((int16 *) height.data)[idx] =
                            (int16) interp_dem(lat[idx],
                            lon[idx], &dem_sds_st);
                }
            }
        }
        
        if (!crefl_in_options_st.TOA) {
            for (irow = 0; irow < sds_st_arr[REFSDS].rowsperscan; irow++) {
                for (jcol = 0; jcol < sds_st_arr[REFSDS].Np; jcol++) {
                    idx = irow * sds_st_arr[REFSDS].Np + jcol;
                    
                    if (process_sensor == SENSOR_MODIS) {
                        if (!modis_test_solz_ne_fillvalue (&sds_st_arr [SOLZ], idx)) {
                            continue;
                        }
                    } else {
                        if (!viirs_test_solz_ne_fillvalue (&sds_st_arr [SOLZ], idx)) {
                            continue;
                        }
                    }
                    
                    if (process_sensor == SENSOR_MODIS) {
                        phi = modis_calculate_phi (&sds_st_arr [SOLA], &sds_st_arr [SENA], idx);
                        muv = modis_calculate_muv (&sds_st_arr[SENZ], idx);
                    } else {
                        phi = viirs_calculate_phi (&sds_st_arr [SOLA], &sds_st_arr [SENA], idx);
                        muv = viirs_calculate_muv (&sds_st_arr[SENZ], idx);
                    }
                    
                    /* Willem.Marais@ssec.wisc.edu - 05/20/2013 - BEGIN */
                    if (process_sensor == SENSOR_MODIS) {
                        if (modis_getatmvariables(mus[idx], muv, phi,
                            ((int16 *) height.data)[idx],
                            band_process, &sphalb[idx * Nbands],
                            &rhoray[idx * Nbands],
                            &TtotraytH2O[idx * Nbands],
                            &tOG[idx * Nbands]) == -1) {
                            
                            modis_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                        }
                    } else {
                        if (viirs_getatmvariables(mus[idx], muv, phi,
                            ((int16 *) height.data)[idx],
                            band_process, &sphalb[idx * Nbands],
                            &rhoray[idx * Nbands],
                            &TtotraytH2O[idx * Nbands],
                            &tOG[idx * Nbands]) == -1) {
                            
                            viirs_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                        }
                    }
                    
                    /*
                    if (getatmvariables(mus[idx], muv, phi,
                        ((int16 *) height.data)[idx],
                        band_process, &sphalb[idx * Nbands],
                        &rhoray[idx * Nbands],
                        &TtotraytH2O[idx * Nbands],
                        &tOG[idx * Nbands]) == -1) {
                        
                        if (process_sensor == SENSOR_MODIS) {
                            modis_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                        } else {
                            viirs_set_solz_to_fillvalue (&sds_st_arr [SOLZ], idx);
                        }
                    }
                    */
                    /* Willem.Marais@ssec.wisc.edu - 05/20/2013 - END */
                }
            }
        }
        
        for (band_idx = 0; band_idx < Nbands; band_idx++) {
            if (!crefl_in_options_st.process[band_idx])
                continue;
            
            ib = sds_st_arr [band_idx].crefl_band_number;
            
            aggfactor =
                ((float)outsds[band_idx].rowsperscan) / ((float)sds_st_arr[REFSDS].rowsperscan);
            for (irow = 0; irow < outsds[band_idx].rowsperscan; irow++) {
                if (!crefl_in_options_st.nearest) {
                    fractrow = (float) irow / aggfactor - 0.5;  /* We want fractrow integer on coarse pixel center */
                    crsrow1 = floor(fractrow);
                    crsrow2 = crsrow1 + 1;
                    if (crsrow1 < 0)
                        crsrow1 = crsrow2 + 1;
                    if (crsrow2 > sds_st_arr[REFSDS].rowsperscan - 1)
                        crsrow2 = crsrow1 - 1;
                    t = (fractrow - crsrow1) / (crsrow2 - crsrow1);
                }

                for (jcol = 0; jcol < outsds[band_idx].Np; jcol++) {
                    idx = irow * outsds[band_idx].Np + jcol;
                    crsidx = (int) (irow / aggfactor) * sds_st_arr[REFSDS].Np + (int) (jcol / aggfactor);
                    
                    /* Willem.Marais@ssec.wisc.edu - Note: The L1B data are checked at this point. Will it be unique 
                       for VIIRS? */
                    /* Bad geolocation or night pixel OR L1B is read as int16, not uint16, so faulty is negative */
                    /* TODO: VIIRS */
                    if (process_sensor == SENSOR_MODIS) {
                        if (modis_clause_solz_neg_l1bdata (&sds_st_arr[SOLZ], &sds_st_arr[band_idx], &outsds [band_idx], crsidx, idx) == 1) {
                            continue;
                        }
                    } else {
                        if (viirs_clause_solz_neg_l1bdata (&sds_st_arr[SOLZ], &sds_st_arr[band_idx], &outsds [band_idx], crsidx, idx) == 1) {
                            continue;
                        }
                    }
                    
                    if (crefl_in_options_st.nearest) {
                        mus0 = mus[crsidx];
                        if (!crefl_in_options_st.TOA) {
                            rhoray0 = rhoray[crsidx * Nbands + ib];
                            sphalb0 = sphalb[crsidx * Nbands + ib];
                            if (sphalb0 <= 0.0F) { /* Atm variables not computed successfully in this band */
                                ((int16 *) outsds[band_idx].data)[idx] = *(int16 *)outsds[band_idx].fillvalue;
                                
                                continue;
                            }
                        }
                    } else {
                        fractcol = ((float) jcol) / aggfactor - 0.5F; /* We want fractcol integer on coarse pixel center */
                        crscol1 = (int) floor(fractcol);
                        crscol2 = crscol1 + 1;
                        if (crscol1 < 0)
                            crscol1 = crscol2 + 1;
                        if (crscol2 > sds_st_arr[REFSDS].Np - 1)
                            crscol2 = crscol1 - 1;
                        u = (fractcol - crscol1) / (crscol2 - crscol1); /* We want u=0 on coarse pixel center */
                        crsidx11 = crsrow1 * sds_st_arr[REFSDS].Np + crscol1;
                        crsidx12 = crsrow1 * sds_st_arr[REFSDS].Np + crscol2;
                        crsidx21 = crsrow2 * sds_st_arr[REFSDS].Np + crscol1;
                        crsidx22 = crsrow2 * sds_st_arr[REFSDS].Np + crscol2;
                        mus0 = t * u * mus[crsidx22] + (1.0F - t) * u * mus[crsidx12] + 
                            t * (1.0F - u) * mus[crsidx21] + (1.0F - t) * (1.0F - u) * mus[crsidx11];
                        
                        if (process_sensor == SENSOR_MODIS) {
                            bad = modis_test_bad_val (&sds_st_arr[SOLZ], crsidx11, crsidx12, crsidx21, crsidx22);
                        } else {
                            bad = viirs_test_bad_val (&sds_st_arr[SOLZ], crsidx11, crsidx12, crsidx21, crsidx22);
                        }
                        
                        if (bad) {
                            ((int16 *) outsds[band_idx].data)[idx] = *(int16 *) outsds[band_idx].fillvalue;

                            continue;
                        }

                        if (!crefl_in_options_st.TOA) {
                            rhoray11 = rhoray[crsidx11 * Nbands + ib];
                            rhoray12 = rhoray[crsidx12 * Nbands + ib];
                            rhoray21 = rhoray[crsidx21 * Nbands + ib];
                            rhoray22 = rhoray[crsidx22 * Nbands + ib];
                            rhoray0 = t * u * rhoray22 + (1.0F - t) * u * rhoray12 + t * (1.0F - u) * rhoray21 + 
                                (1.0F - t) * (1.0F - u) * rhoray11;
                            sphalb11 = sphalb[crsidx11 * Nbands + ib];
                            sphalb12 = sphalb[crsidx12 * Nbands + ib];
                            sphalb21 = sphalb[crsidx21 * Nbands + ib];
                            sphalb22 = sphalb[crsidx22 * Nbands + ib];

                            bad = (sphalb11 <= 0.0F) || (sphalb12 <= 0.0F) || (sphalb21 <= 0.0F) || (sphalb22 <= 0.0F);

                            if (bad) {
                                ((int16 *) outsds[band_idx].data)[idx] = *(int16 *) outsds[band_idx].fillvalue;
                                continue;
                            }
                            sphalb0 = t * u * sphalb22 + (1.0F - t) * u * sphalb12 + 
                                t * (1.0F - u) * sphalb21 + (1.0F - t) * (1.0F - u) * sphalb11;
                        }
                    }

                    /* TOA reflectance */
                    /* Willem.Marais@ssec.wisc.edu - Note: At this point if it is VIIRS data, there 
                       shouldn't be a division of mus0 */
                    /* TODO: casting */
                    /* TODO: VIIRS */
                    if (process_sensor == SENSOR_MODIS) {
                        refl = modis_calculate_refl (&sds_st_arr [band_idx], idx, mus0);
                    } else {
                        refl = viirs_calculate_refl (&sds_st_arr [band_idx], idx);
                    }
                    
                    /* corrected reflectance */
                    if (!crefl_in_options_st.TOA)
                        refl = correctedrefl (refl, TtotraytH2O[crsidx * Nbands + ib], tOG[crsidx * Nbands + ib], rhoray0, sphalb0);
                    
                    /* reflectance bounds checking */
                    if (refl > crefl_in_options_st.reflmax)
                        refl = crefl_in_options_st.reflmax;
                    if (refl < crefl_in_options_st.reflmin)
                        refl = crefl_in_options_st.reflmin;
                    
                    ((int16 *) outsds[band_idx].data)[idx] = (int16) (refl / outsds[band_idx].factor + 0.5);
                }
            }
        } /* for (band_idx = 0; band_idx < Nbands; band_idx++) */
        
        /* write current scan line for all processed bands */
        if (write_scan_outsds(iscan, crefl_in_options_st.process, outsds)) {
            fprintf(stderr, "Cannot write scan %d of SDS %s\n",
                iscan, outsds[band_idx].name);
            exit(1);
        }
        
        for (idx = 0; idx < sds_st_arr[LON].rowsperscan * sds_st_arr[LON].Np; idx++) {
            for (geo_idx = 0; geo_idx < 2; geo_idx++) {
                ((float32 *) geo_outsds [geo_idx].data)[idx] = 
                    ((float32 *) sds_st_arr [geo_idx + LON].data)[idx];
            }
        }
        
        /* TODO: Copy geolocation data. 10/08 Willem.Marais@ssec.wisc.edu */
        /*  Write geolocation data to output file: write_scan */
        write_scan_geo_outsds (iscan, geo_outsds);
        
    } /* END: for (iscan = 0; iscan < Nscans; iscan++) */
    
    if (process_sensor == SENSOR_MODIS) {
        modis_finalize_data (&crefl_in_options_st, sds_st_arr);
    } else {
        viirs_finalize_data (&crefl_in_options_st, sds_st_arr);
    }
    
    /* TODO: Go through sds array and get the file id of these files */
    for (ib = 0; ib < Nbands; ib++) {
        if (crefl_in_options_st.process[ib]) {
            SDendaccess(outsds[ib].id);
            free (outsds[ib].data);
        }
    }
    
    /* Release resources in the geolocation SDS sturctures */
    for (ib = 0; ib < 2; ib++) {
        SDendaccess(geo_outsds[ib].id);
        free(geo_outsds[ib].data);
    }
    
    /* Close output file */
    SDend(sd_id);
    
    free(height.data);
    free(mus);
    
    if (!crefl_in_options_st.TOA) {
        free(tOG);
        free(TtotraytH2O);
        free(sphalb);
        free(rhoray);
    }
    
    dem_finalize_data (&crefl_in_options_st, &dem_sds_st);
    
    return 0;
}

/* set 250-m, 500-m, or 1-km line and sample dimension names for MODIS bands
given number of samples across scan */
void
set_dimnames(int samples, char **dimname1, char **dimname2)
{   
    fprintf (stderr, "samples: %d\n", samples);
    switch (samples) {
    case MODIS_NUMGEOCOLPERSCAN:
        *dimname1 = "lines_5km";
        *dimname2 = "samples_5km";
        break;
        
    case MODIS_NUM1KMCOLPERSCAN:
        *dimname1 = "lines_1km";
        *dimname2 = "samples_1km";
        break;

    case (MODIS_NUM1KMCOLPERSCAN * 2):
        *dimname1 = "lines_500m";
        *dimname2 = "samples_500m";
        break;

    case (MODIS_NUM1KMCOLPERSCAN * 4):
        *dimname1 = "lines_250m";
        *dimname2 = "samples_250m";
        break;

    case VIIRS_NUM750MCOLPERSCAN:
        *dimname1 = "lines_750m";
        *dimname2 = "samples_750m";
        break;
    
    case (VIIRS_NUM750MCOLPERSCAN * 2):
        *dimname1 = "lines_350m";
        *dimname2 = "samples_350m";
        break;
    
    default:
        *dimname1 = *dimname2 = (char *) NULL;
        break;
    }
}

int
interp_dem(float lat, float lon, SDS *dem_sds_st)
{
    float fractrow, fractcol, t, u;
    int demrow1, demcol1, demrow2, demcol2;
    int height11, height12, height21, height22;
    int height;

    fractrow = (90.0F - lat) * dem_sds_st->Nl / 180.0F;
    demrow1 = (int) floorf(fractrow);
    demrow2 = demrow1 + 1;
    if (demrow1 < 0)
        demrow1 = demrow2 + 1;
    if (demrow2 > dem_sds_st->Nl - 1)
        demrow2 = demrow1 - 1;
    t = (fractrow - demrow1) / (demrow2 - demrow1);

    fractcol = (lon + 180.0F) * dem_sds_st->Np / 360.0F;
    demcol1 = (int) floorf(fractcol);
    demcol2 = demcol1 + 1;
    if (demcol1 < 0)
        demcol1 = demcol2 + 1;
    if (demcol2 > dem_sds_st->Np - 1)
        demcol2 = demcol1 - 1;
    u = (fractcol - demcol1) / (demcol2 - demcol1);

    height11 = ((int16 *) dem_sds_st->data)[demrow1 * dem_sds_st->Np + demcol1];
    height12 = ((int16 *) dem_sds_st->data)[demrow1 * dem_sds_st->Np + demcol2];
    height21 = ((int16 *) dem_sds_st->data)[demrow2 * dem_sds_st->Np + demcol1];
    height22 = ((int16 *) dem_sds_st->data)[demrow2 * dem_sds_st->Np + demcol2];
    height = (int) (t * u * height22 + t * (1.0F - u) * height21 +
        (1.0F - t) * u * height12 + (1.0F - t) * (1.0F - u) * height11);

    if (height < 0)
        height = 0;
    return height;
}

/* Write current scan line for all processed bands.
Returns 0 if no errors, non-zero otherwise. */
int write_scan_outsds(int iscan, unsigned char *process, SDS outsds[Nbands])
{
    int ib;

    for (ib = 0; ib < Nbands; ib++) {
        if (!process[ib])
            continue;

        outsds[ib].start[0] = iscan * outsds[ib].rowsperscan;
        if (SDwritedata(outsds[ib].id, outsds[ib].start, NULL,
            outsds[ib].edges, outsds[ib].data) == -1)
            return 1;
    }

    return 0;
}

int write_scan_geo_outsds(int iscan, SDS geo_outsds[2])
{
    int ib;
    
    for (ib = 0; ib < 2; ib++) {
        geo_outsds[ib].start[0] = iscan * geo_outsds[ib].rowsperscan;
        
        if (SDwritedata(geo_outsds[ib].id, geo_outsds[ib].start, NULL, geo_outsds[ib].edges, geo_outsds[ib].data) == -1) {
            return 1;
        }
    }
    
    return 0;
}

float
csalbr(float tau)
{
    return (3.0F * tau - fintexp3(tau) * (4.0F + 2.0F * tau) +
        2.0 * expf(-tau)) / (4.0F + 3.0F * tau);
}

double
fintexp1(float tau)
{
    double xx, xftau;
    int i;
    const double a[6] = { -.57721566, 0.99999193, -0.24991055,
        0.05519968, -0.00976004, 0.00107857
    };
    xx = a[0];
    xftau = 1.0;
    for (i = 1; i < 6; i++) {
        xftau *= tau;
        xx += a[i] * xftau;
    }

    return xx - log(tau);
}

double
fintexp3(float tau)
{
    return (expf(-tau) * (1.0F - tau) + tau * tau * fintexp1(tau)) / 2.0;
}

void
chand(float phi, float muv, float mus, float *taur, float *rhoray,
    float *trup, float *trdown, unsigned char *process)
{
/*
phi: azimuthal difference between sun and observation in degree
     (phi=0 in backscattering direction)
mus: cosine of the sun zenith angle
muv: cosine of the observation zenith angle
taur: molecular optical depth
rhoray: molecular path reflectance
constant xdep: depolarization factor (0.0279)
         xfd = (1-xdep/(2-xdep)) / (1 + 2*xdep/(2-xdep)) = 2 * (1 - xdep) / (2 + xdep) = 0.958725775
*/
    const double xfd = 0.958725775;
    const float xbeta2 = 0.5F;
    float pl[5];
    double fs01, fs02, fs0, fs1, fs2;
    const float as0[10] =
        { 0.33243832F, 0.16285370F, -0.30924818F, -0.10324388F,
        0.11493334F, -6.777104e-02F, 1.577425e-03F, -1.240906e-02F,
        3.241678e-02F, -3.503695e-02F
    };
    const float as1[2] = { 0.19666292F, -5.439061e-02F };
    const float as2[2] = { 0.14545937F, -2.910845e-02F };
    float phios, xcos1, xcos2, xcos3;
    float xph1, xph2, xph3, xitm1, xitm2;
    float xlntaur, xitot1, xitot2, xitot3;
    int i, ib;

    phios = phi + 180.0F;
    xcos1 = 1.0F;
    xcos2 = cos(phios * DEG2RAD);
    xcos3 = cos(2.0 * phios * DEG2RAD);
    xph1 =
        1.0 + (3.0F * mus * mus - 1.0F) * (3.0F * muv * muv -
        1.0F) * xfd / 8.0;
    xph2 =
        -xfd * xbeta2 * 1.5 * mus * muv * sqrtf(1.0F -
        mus * mus) * sqrtf(1.0F - muv * muv);
    xph3 = xfd * xbeta2 * 0.375 * (1.0F - mus * mus) * (1.0F - muv * muv);

    pl[0] = 1.0F;
    pl[1] = mus + muv;
    pl[2] = mus * muv;
    pl[3] = mus * mus + muv * muv;
    pl[4] = mus * mus * muv * muv;

    fs01 = fs02 = 0.0;
    for (i = 0; i < 5; i++) {
        fs01 += (double) (pl[i] * as0[i]);
        fs02 += (double) (pl[i] * as0[5 + i]);
    }

    for (ib = 0; ib < Nbands; ib++) {
        if (process[ib]) {
            xlntaur = logf(taur[ib]);
            fs0 = fs01 + fs02 * xlntaur;
            fs1 = as1[0] + xlntaur * as1[1];
            fs2 = as2[0] + xlntaur * as2[1];
            trdown[ib] = expf(-taur[ib] / mus);
            trup[ib] = expf(-taur[ib] / muv);
            xitm1 =
                (1.0F - trdown[ib] * trup[ib]) / 4.0F / (mus +
                muv);
            xitm2 = (1.0F - trdown[ib]) * (1.0F - trup[ib]);
            xitot1 = xph1 * (xitm1 + xitm2 * fs0);
            xitot2 = xph2 * (xitm1 + xitm2 * fs1);
            xitot3 = xph3 * (xitm1 + xitm2 * fs2);
            rhoray[ib] =
                xitot1 * xcos1 + xitot2 * xcos2 * 2.0F +
                xitot3 * xcos3 * 2.0F;
        }
    }
}

/* Willem.Marais@ssec.wisc.edu - 05/20/2013 - BEGIN */
int
modis_getatmvariables(float mus, float muv, float phi, int16 height,
    unsigned char *process, float *sphalb, float *rhoray,
    float *TtotraytH2O, float *tOG)
{
    double m, Ttotrayu, Ttotrayd, tO3, tO2, tH2O;
    float psurfratio;
    int j, ib;
    
    /* Original coefficients */
    /*
    Values for bands 9-16 below provided by B Murch and C Hu Univ South Florida
    IMaRS, obtained from SEADAS.
    For the moment I've retained the Jacques values for 1-8 but show the differing
    SEADAS values in the commented out line.
    */
    const float aH2O[Nbands] =
        { -5.60723, -5.25251, 0, 0, -6.29824, -7.70944, -3.91877, 0, 0, 0,
        0, 0, 0, 0, 0, 0
    };
    const float bH2O[Nbands] =
        { 0.820175, 0.725159, 0, 0, 0.865732, 0.966947, 0.745342, 0, 0, 0,
        0, 0, 0, 0, 0, 0
    };
    /*const float aO3[Nbands]={ 0.0711,    0.00313, 0.0104,     0.0930,   0, 0, 0, 0.00244, 0.00383, 0.0225, 0.0663, 0.0836, 0.0485, 0.0395, 0.0119, 0.00263};*/
    const float aO3[Nbands] =
        { 0.0715289, 0, 0.00743232, 0.089691, 0, 0, 0, 0.001, 0.00383,
        0.0225, 0.0663, 0.0836, 0.0485, 0.0395, 0.0119, 0.00263
    };
    /*const float taur0[Nbands] = { 0.0507,  0.0164,  0.1915,  0.0948,  0.0036,  0.0012,  0.0004,  0.3109, 0.2375, 0.1596, 0.1131, 0.0994, 0.0446, 0.0416, 0.0286, 0.0155};*/
    const float taur0[Nbands] =
        { 0.05100, 0.01631, 0.19325, 0.09536, 0.00366, 0.00123, 0.00043,
        0.3139, 0.2375, 0.1596, 0.1131, 0.0994, 0.0446, 0.0416, 0.0286,
            0.0155
    };
    
    float taur[Nbands], trup[Nbands], trdown[Nbands];
    static float sphalb0[MAXNUMSPHALBVALUES];
    static char first_time = TRUE;

    if (first_time) {
        sphalb0[0] = 0.0F;
        for (j = 1; j < MAXNUMSPHALBVALUES; j++)
            sphalb0[j] = csalbr(j * TAUSTEP4SPHALB);
        first_time = FALSE;
    }

    m = 1.0 / mus + 1.0 / muv;
    if (m > MAXAIRMASS)
        return -1;

    psurfratio = expf(-height / (float) SCALEHEIGHT);
    for (ib = 0; ib < Nbands; ib++)
        if (process[ib])
            taur[ib] = taur0[ib] * psurfratio;
    
    chand(phi, muv, mus, taur, rhoray, trup, trdown, process);
    
    for (ib = 0; ib < Nbands; ib++) {
        if (!process[ib])
            continue;
        
        if (taur[ib] / TAUSTEP4SPHALB >= MAXNUMSPHALBVALUES) {
            sphalb[ib] = -1.0F;
            /* Use sphalb as flag to indicate atm variables are not computed successfully */
            continue;
        }
        
        sphalb[ib] = sphalb0[(int) (taur[ib] / TAUSTEP4SPHALB + 0.5)];
        Ttotrayu = ((2 / 3. + muv) + (2 / 3. - muv) * trup[ib]) / (4 / 3. + taur[ib]);
        Ttotrayd = ((2 / 3. + mus) + (2 / 3. - mus) * trdown[ib]) / (4 / 3. + taur[ib]); 
        tO3 = tO2 = tH2O = 1.0;
        if (aO3[ib] != 0) tO3 = exp(-m * UO3 * aO3[ib]);
        if (bH2O[ib] != 0) tH2O = exp(-exp(aH2O[ib] + bH2O[ib] * log(m * UH2O)));
        
        /*
        t02 = exp(-m * aO2);
        */
        
        TtotraytH2O[ib] = Ttotrayu * Ttotrayd * tH2O;
        tOG[ib] = tO3 * tO2;
    }

    return 0;
}

/*
    The following code was taken from the file cviirs3.1.c. Refer to the following link to get the 
    VIIRS crefl code:
    https://forums.ssec.wisc.edu/viewtopic.php?f=42&t=304
*/
int
viirs_getatmvariables(float mus, float muv, float phi, int16 height,
    unsigned char *process, float *sphalb, float *rhoray,
    float *TtotraytH2O, float *tOG)
{
    double m, Ttotrayu, Ttotrayd, tO3, tO2, tH2O;
    float psurfratio;
    int j, ib;
    
    const float aH2O[Nbands]={0.000406601 ,0.0015933 , 0,1.78644e-05 ,0.00296457 ,0.000617252 , 
                              0.000996563,0.00222253 ,0.00094005 , 0.000563288, 0, 0, 0, 0, 0, 0 };
    const float bH2O[Nbands]={ 0.812659,0.832931 , 1., 0.8677850, 0.806816 , 0.944958 ,0.78812 ,
                               0.791204 ,0.900564 ,0.942907 , 0, 0, 0, 0, 0, 0 };
    const float aO3[Nbands]={ 0.0433461, 0.0,    0.0178299   ,0.0853012 , 0, 0, 0, 0.0813531,   
                              0, 0, 0.0663, 0.0836, 0.0485, 0.0395, 0.0119, 0.00263};
    const float taur0[Nbands] = { 0.04350, 0.01582, 0.16176, 0.09740,0.00369 ,0.00132 ,0.00033 ,
                                  0.05373 ,0.01561 ,0.00129, 0.1131, 0.0994, 0.0446, 0.0416, 0.0286, 0.0155};
    
    float taur[Nbands], trup[Nbands], trdown[Nbands];
    static float sphalb0[MAXNUMSPHALBVALUES];
    static char first_time = TRUE;

    if (first_time) {
        sphalb0[0] = 0.0F;
        for (j = 1; j < MAXNUMSPHALBVALUES; j++)
            sphalb0[j] = csalbr(j * TAUSTEP4SPHALB);
        first_time = FALSE;
    }
    
    m = 1.0 / mus + 1.0 / muv;
    if (m > MAXAIRMASS)
        return -1;
    
    psurfratio = expf(-height / (float) SCALEHEIGHT);
    for (ib = 0; ib < Nbands; ib++)
        if (process[ib])
            taur[ib] = taur0[ib] * psurfratio;
    
    chand(phi, muv, mus, taur, rhoray, trup, trdown, process);
    
    for (ib = 0; ib < Nbands; ib++) {
        if (!process[ib]) continue;

        if (taur[ib] / TAUSTEP4SPHALB >= MAXNUMSPHALBVALUES) {
            sphalb[ib] = -1.0F;
            /* Use sphalb as flag to indicate atm variables are not computed successfully */
            continue;
            }

        sphalb[ib] = sphalb0[(int)(taur[ib] / TAUSTEP4SPHALB + 0.5)];
        Ttotrayu = ((2 / 3. + muv) + (2 / 3. - muv) * trup[ib])   / (4 / 3. + taur[ib]);
        Ttotrayd = ((2 / 3. + mus) + (2 / 3. - mus) * trdown[ib]) / (4 / 3. + taur[ib]);
        tO3 = tO2 = tH2O = 1.0;
        if (aO3[ib] != 0) tO3 = exp(-m * UO3 * aO3[ib]);
        if (bH2O[ib] != 0) tH2O = exp(-(aH2O[ib]*(pow((m * UH2O),bH2O[ib]))));
/*
      t02 = exp(-m * aO2);
*/
        TtotraytH2O[ib] = Ttotrayu * Ttotrayd * tH2O;
        tOG[ib] = tO3 * tO2;
    }
    
    return 0;
}
/* Willem.Marais@ssec.wisc.edu - 05/20/2013 - END */

float correctedrefl(float refl, float TtotraytH2O, float tOG, float rhoray, float sphalb)
{
    float corr_refl;

    corr_refl = (refl / tOG - rhoray) / TtotraytH2O;
    corr_refl /= (1.0F + corr_refl * sphalb);
    return corr_refl;
}

int
init_output_sds(int32 sd_id, unsigned char *process,
    SDS outsds[Nbands], SDS sds_st_arr[Nitems], int gzip, int verbose)
{
    int ib;
    int32 dim_id;
    char *dimname1, *dimname2;

    HDF_CHUNK_DEF chunk_def;

    /* same fill value will be used for all output SDSs */
    static int16 fillvalue = FILL_INT16;

    /* band naming convention will be "CorrRefl_XX"
     * (11 characters + terminating null) */
    char name[16];

    /* write SDS-specific attributes and dimension names */
    for (ib = 0; ib < Nbands; ib++) {
        if (!process[ib])
            continue;
        
        outsds [ib].crefl_band_number = ib;
        outsds [ib].sensor_band_number = sds_st_arr [ib].sensor_band_number;
        
        outsds[ib].num_type = DFNT_INT16;
        outsds[ib].factor = 0.0001;
        outsds[ib].offset = 0;
        outsds[ib].rank = 2;
        
        /* Outout array names */
        sprintf(name, "CorrRefl_%2.2d", sds_st_arr [ib].sensor_band_number + 1);
        if (!(outsds[ib].name = strdup(name)))
            return 1;

        outsds[ib].Nl = outsds[ib].dim_sizes[0] = sds_st_arr[ib].Nl;
        outsds[ib].Np = outsds[ib].dim_sizes[1] = sds_st_arr[ib].Np;
        outsds[ib].rowsperscan = sds_st_arr[ib].rowsperscan;
        if (verbose)
            printf("Creating SDS %s: %dx%d\n", outsds[ib].name,
                outsds[ib].Np, outsds[ib].Nl);
        if ((outsds[ib].id =
            SDcreate(sd_id, outsds[ib].name, outsds[ib].num_type,
                outsds[ib].rank, outsds[ib].dim_sizes)) == -1) {
            fprintf(stderr, "Cannot create SDS %s\n",
                outsds[ib].name);
            return 1;
        }
        
        outsds[ib].fillvalue = &fillvalue;
        if (SDsetfillvalue(outsds[ib].id, outsds[ib].fillvalue)) {
            fprintf(stderr, "Cannot write fill value of SDS %s\n",
                outsds[ib].name);
            return 1;
        }
        if (SDsetattr
            (outsds[ib].id, "scale_factor", DFNT_FLOAT64, 1,
            &outsds[ib].factor) == -1
            || SDsetattr(outsds[ib].id, "add_offset", DFNT_FLOAT64, 1,
            &outsds[ib].offset) == -1) {
            fprintf(stderr,
                "Cannot write scale factor and offset of SDS \"%s\"\n",
                outsds[ib].name);
            return 1;
        }
        if (SDsetattr(outsds[ib].id, "units", DFNT_CHAR8, 4,
            "none") == -1) {
            fprintf(stderr,
                "Cannot write units attribute of SDS \"%s\"\n",
                outsds[ib].name);
            return 1;
        }
        
        /* set dimensions */
        outsds[ib].start[1] = 0;
        outsds[ib].edges[0] = outsds[ib].rowsperscan;
        outsds[ib].edges[1] = outsds[ib].Np;

        /* allocate memory for band output data */
        outsds[ib].data =
            malloc(outsds[ib].rowsperscan * outsds[ib].Np *
            DFKNTsize(outsds[ib].num_type));
        if (!outsds[ib].data) {
            fputs("Error allocating memory.\n", stderr);
            return 1;
        }
        
        /* set optional compression */
        if (gzip) {
            chunk_def.chunk_lengths[0] =
                chunk_def.comp.chunk_lengths[0] =
                outsds[ib].edges[0];
            chunk_def.chunk_lengths[1] =
                chunk_def.comp.chunk_lengths[1] =
                outsds[ib].edges[1];
            chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
            chunk_def.comp.cinfo.deflate.level = 4;
            if (SDsetchunk(outsds[ib].id, chunk_def,
                HDF_CHUNK | HDF_COMP)
                == FAIL) {
                fprintf(stderr,
                    "Cannot set chunks for SDS %s\n",
                    outsds[ib].name);
                return 1;
            }
        }
        
        set_dimnames(outsds[ib].Np, &dimname1, &dimname2);
        
        if (verbose)
            printf("(%s x %s)\n", dimname1, dimname2);

        /* dimension names */
        if ((dim_id = SDgetdimid(outsds[ib].id, 0)) == -1) {
            fputs("Error getting dimension ID1.\n", stderr);
            return 1;
        }
        
        if (SDsetdimname(dim_id, dimname1) == -1) {
            fprintf(stderr,
                "Cannot set first dimension name for SDS %s\n",
                outsds[ib].name);
            return 1;
        }
        
        if ((dim_id = SDgetdimid(outsds[ib].id, 1)) == -1) {
            fputs("Error getting dimension ID2.\n", stderr);
            return 1;
        }
        
        if (SDsetdimname(dim_id, dimname2) == -1) {
            fprintf(stderr,
                "Cannot set second dimension name for SDS %s\n",
                outsds[ib].name);
            return 1;
        }
    }

    return 0;
}

/* Willem.Marais@ssec.wisc.edu - 10/09/2012 - BEGIN */
int
init_geo_output_sds(int32 sd_id, SDS geo_outsds[2], SDS sds_st_arr[Nitems], int gzip, int verbose)
{
    int ib;
    int32 dim_id;
    char *dimname1, *dimname2;
    static char* sds_names_str [2] = {"Longitude", "Latitude"};
    
    HDF_CHUNK_DEF chunk_def;
    
    /* same fill value will be used for all output SDSs */
    static float fillvalue = FILL_FLOAT32;
    
    /* write SDS-specific attributes and dimension names */
    for (ib = 0; ib < 2; ib++) {
        /* Willem.Marais@ssec.wisc.edu - 10/09/2012 - Are these necessary? */
        geo_outsds [ib].crefl_band_number = ib;
        geo_outsds [ib].sensor_band_number = sds_st_arr [ib + LON].sensor_band_number;
        
        geo_outsds[ib].num_type = DFNT_FLOAT32;
        geo_outsds[ib].factor = 1.0;
        geo_outsds[ib].offset = 0;
        geo_outsds[ib].rank = 2;
        
        if (!(geo_outsds[ib].name = strdup(sds_names_str [ib])))
            return 1;
        
        geo_outsds[ib].Nl = geo_outsds[ib].dim_sizes[0] = sds_st_arr[ib + LON].Nl;
        geo_outsds[ib].Np = geo_outsds[ib].dim_sizes[1] = sds_st_arr[ib + LON].Np;
        geo_outsds[ib].rowsperscan = sds_st_arr[ib + LON].rowsperscan;
        if (verbose)
            printf("Creating SDS %s: %dx%d\n", geo_outsds[ib].name,
                geo_outsds[ib].Np, geo_outsds[ib].Nl);
        if ((geo_outsds[ib].id =
            SDcreate(sd_id, geo_outsds[ib].name, geo_outsds[ib].num_type,
                geo_outsds[ib].rank, geo_outsds[ib].dim_sizes)) == -1) {
            fprintf(stderr, "Cannot create SDS %s\n",
                geo_outsds[ib].name);
            return 1;
        }
        
        geo_outsds[ib].fillvalue = &fillvalue;
        if (SDsetfillvalue(geo_outsds[ib].id, geo_outsds[ib].fillvalue)) {
            fprintf(stderr, "Cannot write fill value of SDS %s\n",
                geo_outsds[ib].name);
            return 1;
        }
        if (SDsetattr
            (geo_outsds[ib].id, "scale_factor", DFNT_FLOAT64, 1,
            &geo_outsds[ib].factor) == -1
            || SDsetattr(geo_outsds[ib].id, "add_offset", DFNT_FLOAT64, 1,
            &geo_outsds[ib].offset) == -1) {
            fprintf(stderr,
                "Cannot write scale factor and offset of SDS \"%s\"\n",
                geo_outsds[ib].name);
            return 1;
        }
        if (SDsetattr(geo_outsds[ib].id, "units", DFNT_CHAR8, 4,
            "none") == -1) {
            fprintf(stderr,
                "Cannot write units attribute of SDS \"%s\"\n",
                geo_outsds[ib].name);
            return 1;
        }
        
        /* set dimensions */
        geo_outsds[ib].start[1] = 0;
        geo_outsds[ib].edges[0] = geo_outsds[ib].rowsperscan;
        geo_outsds[ib].edges[1] = geo_outsds[ib].Np;

        /* allocate memory for band output data */
        geo_outsds[ib].data =
            malloc(geo_outsds[ib].rowsperscan * geo_outsds[ib].Np *
            DFKNTsize(geo_outsds[ib].num_type));
        if (!geo_outsds[ib].data) {
            fputs("Error allocating memory.\n", stderr);
            return 1;
        }
        
        /* set optional compression */
        if (gzip) {
            chunk_def.chunk_lengths[0] =
                chunk_def.comp.chunk_lengths[0] =
                geo_outsds[ib].edges[0];
            chunk_def.chunk_lengths[1] =
                chunk_def.comp.chunk_lengths[1] =
                geo_outsds[ib].edges[1];
            chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
            chunk_def.comp.cinfo.deflate.level = 4;
            if (SDsetchunk(geo_outsds[ib].id, chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
                fprintf(stderr, "Cannot set chunks for SDS %s\n", geo_outsds[ib].name);
                return 1;
            }
        }
        
        set_dimnames(geo_outsds[ib].Np, &dimname1, &dimname2);
        
        if (verbose) {
            printf("(%s x %s)\n", dimname1, dimname2);
        }

        /* dimension names */
        if ((dim_id = SDgetdimid(geo_outsds[ib].id, 0)) == -1) {
            fputs("Error getting dimension ID1.\n", stderr);
            return 1;
        }
        
        if (SDsetdimname(dim_id, dimname1) == -1) {
            fprintf(stderr,
                "Cannot set first dimension name for SDS %s\n",
                geo_outsds[ib].name);
            return 1;
        }
        
        if ((dim_id = SDgetdimid(geo_outsds[ib].id, 1)) == -1) {
            fputs("Error getting dimension ID2.\n", stderr);
            return 1;
        }
        
        if (SDsetdimname(dim_id, dimname2) == -1) {
            fprintf(stderr,
                "Cannot set second dimension name for SDS %s\n",
                geo_outsds[ib].name);
            return 1;
        }
    }

    return 0;
}
/* Willem.Marais@ssec.wisc.edu - 10/09/2012 - END */

/*
Write global attributes to HDF output file.
Returns 0 if no errors, non-zero otherwise.

TODO: Adapt for VIIRS
*/
int
write_global_attributes(int32 sd_id, char *MOD021KMfile,
    char *MOD02HKMfile, char *MOD02QKMfile,
    float maxsolz, int sealevel, int TOA, int nearest)
{
    char *ptr;
    int j;
    float32 f32;
    uint8 u8;

    ptr = PROCESS_VERSION_NUMBER;
    j = strlen(ptr);
    if (SDsetattr(sd_id, "ProcessVersionNumber", DFNT_CHAR8, j, ptr))
        return 1;

    if (MOD021KMfile) {
        j = strlen(MOD021KMfile);
        if (SDsetattr
            (sd_id, "1km_input_file", DFNT_CHAR8, j, MOD021KMfile))
            return 1;
    }

    if (MOD02HKMfile) {
        j = strlen(MOD02HKMfile);
        if (SDsetattr
            (sd_id, "500m_input_file", DFNT_CHAR8, j, MOD02HKMfile))
            return 1;
    }

    if (MOD02QKMfile) {
        j = strlen(MOD02QKMfile);
        if (SDsetattr
            (sd_id, "250m_input_file", DFNT_CHAR8, j, MOD02QKMfile))
            return 1;
    }

    f32 = maxsolz;
    if (SDsetattr(sd_id, "MaxSolarZenithAngle", DFNT_FLOAT32, 1, &f32))
        return 1;

    u8 = (uint8) sealevel;
    if (SDsetattr(sd_id, "sealevel", DFNT_UINT8, 1, &u8))
        return 1;
    u8 = (uint8) TOA;
    if (SDsetattr(sd_id, "toa", DFNT_UINT8, 1, &u8))
        return 1;
    u8 = (uint8) nearest;
    if (SDsetattr(sd_id, "nearest", DFNT_UINT8, 1, &u8))
        return 1;
    
    return 0;
}

/* Willem.Marais@ssec.wisc.edu - New functions - BEGIN */
char sensor_name_parm (char *option_str) {
    char return_value = SENSOR_UNKOWN;
    
    if (strcmp (sensor_names [SENSOR_MODIS], option_str) == 0) {
        return_value = SENSOR_MODIS;
    }
    
    if (strcmp (sensor_names [SENSOR_VIIRS], option_str) == 0) {
        return_value = SENSOR_VIIRS;
    }

    return return_value;
}
/* Willem.Marais@ssec.wisc.edu - New functions - END */

